@@ -36,6 +36,7 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <sys/mman.h>
 
 #include <q-lib.h>
 
@@ -43,12 +44,6 @@
 #include "call-counts.h"
 #include "time-hist.h"
 
-/* It is currently not possible to add events incrementally.  Because
-   of that, we need a gross hack in setup_code_sampling() so libpfm
-   knows that PMD4 is already taken for BRANCH_EVENT when we dispatch
-   the code-sampling event (normally CPU_CYCLES).  */
-#define GROSS_HACK	1
-
 #define MIN_POLL_TIMEOUT	 100	/* 100 milliseconds */
 #define MAX_POLL_TIMEOUT	1000	/* 1 second */
 
@@ -91,17 +86,17 @@ static int poll_timeout = MIN_POLL_TIMEOUT;
 static int num_cpus, my_cpu = -1;
 /* PMD register number that's couting BRANCH_EVENT: */
 static int branch_event_pmd = -1;
-static pfarg_reg_t *branch_event_pd, *cs_pd;
+static pfarg_pmd_t *branch_event_pd, *cs_pd;
 /* ditto for code-sampling: */
 static int cs_event_pmd = -1;
 static int duration;
 static char *cs_event_name;
 static int num_pds;
-static pfarg_reg_t pd[PFMLIB_MAX_PMDS];
+static pfarg_pmd_t pd[PFMLIB_MAX_PMDS];
 static struct timeval now, last_time;
 static sig_atomic_t done;
 static unsigned long cs_interval_sum, cs_num_samples;
-static pfm_default_smpl_hdr_t *sampling_buffer;
+static pfm_dfl_smpl_hdr_t *sampling_buffer;
 static int sampling_fd;
 static struct qsys_btb_support *qsys;
 
@@ -141,10 +136,6 @@ fls (unsigned long val)
   return pos;
 }
 
-#ifndef SYS_sched_setaffinity
-# define SYS_sched_setaffinity 1231
-#endif
-
 /* Some versions of libc declare sched_setaffinity() with two args,
    others with 3 args.  Circumvent the madness via syscall()...  */
 
@@ -418,10 +409,9 @@ check_addr_space_mapping (struct addr_space *as, unsigned long addr)
 }
 
 static char *
-process_code_sample (struct addr_space *as,
-		     pfm_default_smpl_entry_t *ent, char *pos)
+process_code_sample (struct addr_space *as, pfm_dfl_smpl_entry_t *ent, char *pos)
 {
-  unsigned long interval = -ent->last_reset_val;
+  uint64_t interval = -ent->last_reset_val;
   uint64_t ip;
 
   cs_interval_sum += interval;
@@ -438,9 +428,9 @@ process_code_sample (struct addr_space *as,
 }
 
 static int
-adjust_rate (pfarg_reg_t *cdesc, double current_rate, double goal_rate)
+adjust_rate (pfarg_pmd_t *cdesc, double current_rate, double goal_rate)
 {
-  unsigned long counter, range_mask;
+  uint64_t counter, range_mask;
 
   if (0.5*goal_rate <= current_rate && current_rate <= 2.0*goal_rate)
     /* No need to adjust rate.  */
@@ -509,8 +499,8 @@ adjust_rate (pfarg_reg_t *cdesc, double current_rate, double goal_rate)
 static void
 process_pmu_samples (void)
 {
-  pfm_default_smpl_hdr_t *hdr = sampling_buffer;
-  pfm_default_smpl_entry_t *ent;
+  pfm_dfl_smpl_hdr_t *hdr = sampling_buffer;
+  pfm_dfl_smpl_entry_t *ent;
   unsigned long i, num_btb_samples = 0, num_code_samples = 0;
   struct addr_space *as = NULL;
   int need_adjustment;
@@ -526,7 +516,7 @@ process_pmu_samples (void)
 
   for (i = 0; i < hdr->hdr_count; ++i)
     {
-      ent = (pfm_default_smpl_entry_t *) pos;
+      ent = (pfm_dfl_smpl_entry_t *) pos;
       pos = old_pos = (char *) (ent + 1);
 
 #if 0
@@ -570,8 +560,8 @@ process_pmu_samples (void)
     need_adjustment |= adjust_rate (cs_pd, num_code_samples / delta,
 				    code_sample_rate);
   if (need_adjustment
-      && perfmonctl (sampling_fd, PFM_WRITE_PMDS, pd, num_pds) < 0)
-    panic ("perfmonctl(PFM_WRITE_PMDS) failed (%s)\n", strerror (errno));
+      && pfm_write_pmds(sampling_fd, pd, num_pds) < 0)
+    panic ("pfm_write_pmds() failed (%s)\n", strerror (errno));
 }
 
 static void
@@ -593,7 +583,7 @@ time_hist_write (void *arg, uintptr_t addr, size_t len, unsigned long *counts)
       if (counts[i] == 0)
 	continue;
 
-      fprintf (fp, "0x%lx %lu\n", addr + (i << INSN_SHIFT), counts[i]);
+      fprintf (fp, "0x%lx %lu\n", (unsigned long)addr + (i << INSN_SHIFT), counts[i]);
     }
   return 0;
 }
@@ -603,7 +593,7 @@ call_count_write (void *arg, uintptr_t from, uintptr_t to, size_t count)
 {
   FILE *fp = arg;
 
-  fprintf (fp, "0x%lx 0x%lx %lu\n", from, to, count);
+  fprintf (fp, "0x%lx 0x%lx %zu\n", (unsigned long)from, (unsigned long)to, count);
   return 0;
 }
 
@@ -779,69 +769,65 @@ write_profile (struct addr_space *as, const char *kallsyms)
 
 
 static void
-setup_code_sampling (pfarg_reg_t *pc, int *num_pcsp,
-		     pfarg_reg_t *pd, int *num_pdsp,
+setup_code_sampling (pfarg_pmc_t *pc, int *num_pcsp,
+		     pfarg_pmd_t *pd, int *num_pdsp,
 		     unsigned int privilege_level_mask)
 {
-  int j, ret, num_pcs = *num_pcsp, num_pds = *num_pdsp;
+  int i, j, ret, num_pcs = *num_pcsp, num_pds = *num_pdsp;
   pfmlib_input_param_t iparam;
   pfmlib_output_param_t oparam;
-  unsigned long reset_value;
-  unsigned long i;
+  uint64_t reset_value;
 
   memset (&iparam, 0, sizeof (iparam));
 
   iparam.pfp_dfl_plm = privilege_level_mask;
   iparam.pfp_event_count = 1;
+
   /* for system-wide monitoring we must use privileged monitors: */
   iparam.pfp_flags = PFMLIB_PFP_SYSTEMWIDE;
 
-#if GROSS_HACK
-  ++iparam.pfp_event_count;
+  /*
+   * no need for GROSS_HACK anymore. We leverage the unavailable PMC
+   * register mechanism.
+   */
+  for(i=0; i < num_pcs; i++)
+    pfm_regmask_set(&iparam.pfp_unavail_pmcs, pc[i].reg_num);
 
   iparam.pfp_events[0].event = cs_event_code;
-  iparam.pfp_events[1].event = cs_event_code;
-#else
-  iparam.pfp_events[0].event = cs_event_code;
-#endif
 
   memset (&oparam, 0, sizeof (oparam));
+
   ret = pfm_dispatch_events (&iparam, NULL, &oparam, NULL);
   if (ret != PFMLIB_SUCCESS)
     panic ("pfm_dispatch_events(): %s\n", pfm_strerror (ret));
 
-#if GROSS_HACK
-  oparam.pfp_pmc_count = 1;
-  oparam.pfp_pmcs[0] = oparam.pfp_pmcs[1];
-#endif
-
-  /* Now setup the PMC and PMD descriptors: */
+  /* Now setup the PMC descriptors: */
 
   for (i = 0; i < oparam.pfp_pmc_count; ++i)
     {
-      for (j = 0; j < num_pcs; ++j)
-	{
-	  if (oparam.pfp_pmcs[i].reg_num == pc[j].reg_num)
-	    panic ("%s: PMC%d is already busy!\n",
-		   __FUNCTION__, pc[j].reg_num);
-	}
       pc[num_pcs + i].reg_num = oparam.pfp_pmcs[i].reg_num;
       pc[num_pcs + i].reg_value = oparam.pfp_pmcs[i].reg_value;
     }
 
-  /* set the PMD reg # for the code-sampling event */
-  pd[num_pds].reg_num = pc[num_pcs].reg_num;
+  /* Now setup the PMD descriptors */
+  for (i=0, j=0; i < iparam.pfp_event_count; i++)
+    {
+      pd[num_pds + i].reg_num = oparam.pfp_pmcs[j].reg_pmd_num;
+      for(; j < oparam.pfp_pmc_count; j++)
+        if (oparam.pfp_pmcs[j].reg_evt_idx != i) break;
+    }
+
   cs_pd = pd + num_pds;
   cs_event_pmd = pd[num_pds].reg_num;
 
   /* Specify what happens when the code sampling event-counter wraps-around: */
 
-  pc[num_pcs].reg_smpl_pmds[0] = 0;		/* don't sample any PMDs */
-  pc[num_pcs].reg_flags |= PFM_REGFL_OVFL_NOTIFY; /* notify on overflow */
-  pc[num_pcs].reg_flags |= PFM_REGFL_RANDOM;	/* randomize the interval */
+  pd[num_pds].reg_smpl_pmds[0] = 0;		/* don't sample any PMDs */
+  pd[num_pds].reg_flags |= PFM_REGFL_OVFL_NOTIFY; /* notify on overflow */
+  pd[num_pds].reg_flags |= PFM_REGFL_RANDOM;	/* randomize the interval */
 
   if (cs_event_code == cs_cycle_code)
-    reset_value = - (long) (cycle_frequency / code_sample_rate);
+    reset_value = - (uint64_t) (cycle_frequency / code_sample_rate);
   else
     reset_value = -100000;	/* your guess is as good as mine... */
 
@@ -856,10 +842,6 @@ setup_code_sampling (pfarg_reg_t *pc, int *num_pcsp,
   *num_pdsp += 1;
 }
 
-
-
-
-
 unsigned long
 get_cycle_frequency (void)
 {
@@ -896,7 +878,7 @@ init_qsys_btb_support(void)
       q++;
   }
   if (monitor_interruptions)
-    panic("PMU model does not support interruption monitoring, non Branch Trace Buffer\n");
+    panic("PMU model does not support interruption monitoring, no Branch Trace Buffer\n");
   return;
 found:
   qsys = *q;
@@ -937,8 +919,9 @@ init_default_cycle_event(void)
 static void
 profile_cpu (int cpu)
 {
-  pfm_default_smpl_ctx_arg_t context;
-  pfarg_reg_t pc[PFMLIB_MAX_PMCS];
+  pfarg_ctx_t context;
+  pfm_dfl_smpl_arg_t buf_arg;
+  pfarg_pmc_t pc[PFMLIB_MAX_PMCS];
   int num_pcs, ret, buffer_full;
   size_t sampling_buffer_size;
   pfmlib_options_t options;
@@ -959,7 +942,8 @@ profile_cpu (int cpu)
 
   memset (&options, 0, sizeof (options));
   options.pfm_debug = 0;	/* set to 1 for debug */
-  options.pfm_verbose = 0;	/* set to 1 for debug */
+  if (verbose)
+    options.pfm_verbose = 1;	/* set to 1 for verbose */
   pfm_set_options (&options);
 
   num_pcs = num_pds = 0;
@@ -1020,37 +1004,39 @@ profile_cpu (int cpu)
   /* Create the perfmon context: */
 
   memset (&context, 0, sizeof (context));
-  memcpy (context.ctx_arg.ctx_smpl_buf_id, (pfm_uuid_t) PFM_DEFAULT_SMPL_UUID,
-	  sizeof (pfm_uuid_t));
-  context.buf_arg.buf_size = sampling_buffer_size;
-  context.ctx_arg.ctx_flags = PFM_FL_SYSTEM_WIDE;
-  if (perfmonctl (0, PFM_CREATE_CONTEXT, &context, 1) == -1)
+  memset (&buf_arg, 0, sizeof (buf_arg));
+  memcpy (context.ctx_smpl_buf_id, (pfm_uuid_t) PFM_DFL_SMPL_UUID, sizeof (pfm_uuid_t));
+  buf_arg.buf_size = sampling_buffer_size;
+  context.ctx_flags = PFM_FL_SYSTEM_WIDE;
+  if (pfm_create_context(&context, &buf_arg, sizeof(buf_arg)) == -1)
     {
       if (errno == ENOSYS)
 	panic ("The kernel doesn't have perfmon enabled.\n");
       else
-	panic ("perfmonctl(PFM_CREATE_CONTEXT) failed (%s)\n",
+	panic ("pfm_create_context() failed (%s)\n",
 	       strerror (errno));
     }
-  sampling_buffer = context.ctx_arg.ctx_smpl_vaddr;
-  sampling_fd = context.ctx_arg.ctx_fd;
+  sampling_fd = context.ctx_fd;
 
-  if (PFM_VERSION_MAJOR(sampling_buffer->hdr_version) != PFM_VERSION_MAJOR(PFM_DEFAULT_SMPL_VERSION))
+  sampling_buffer = mmap(NULL, (size_t)context.ctx_smpl_buf_size, PROT_READ, MAP_PRIVATE, sampling_fd, 0);
+  if (sampling_buffer == MAP_FAILED)
+		panic("cannot mmap sampling buffer errno %s\n", strerror(errno));
+
+  if (PFM_VERSION_MAJOR(sampling_buffer->hdr_version) != PFM_VERSION_MAJOR(PFM_DFL_SMPL_VERSION))
     panic ("kernel default v%u.%u sampling format is not supported\n",
 	   PFM_VERSION_MAJOR (sampling_buffer->hdr_version),
 	   PFM_VERSION_MINOR (sampling_buffer->hdr_version));
 
 
-  if (perfmonctl (sampling_fd, PFM_WRITE_PMCS, pc, num_pcs) < 0)
-    panic ("perfmonctl(PFM_WRITE_PMCS) failed (%s)\n", strerror (errno));
-  if (perfmonctl (sampling_fd, PFM_WRITE_PMDS, pd, num_pds) < 0)
-    panic ("perfmonctl(PFM_WRITE_PMDS) failed (%s)\n", strerror (errno));
-
+  if (pfm_write_pmcs(sampling_fd, pc, num_pcs) < 0)
+    panic ("pfm_write_pmcs() failed (%s)\n", strerror (errno));
+  if (pfm_write_pmds(sampling_fd, pd, num_pds) < 0)
+    panic ("pfm_write_pmds() failed (%s)\n", strerror (errno));
 
   memset (&load_args, 0, sizeof (load_args));
-  load_args.load_pid = getpid ();
-  if (perfmonctl (sampling_fd, PFM_LOAD_CONTEXT, &load_args, 1) == -1)
-    panic ("perfmonctl(PFM_LOAD_CONTEXT) failed (%s)\n", strerror (errno));
+  /* load onto current CPU */
+  if (pfm_load_context(sampling_fd, &load_args) == -1)
+    panic ("pfm_load_context() failed (%s)\n", strerror (errno));
 
   pfd.fd = sampling_fd;
   pfd.events = POLLIN;
@@ -1060,8 +1046,8 @@ profile_cpu (int cpu)
 
   /* Start monitoring. We must go to the kernel because psr.pp cannot
      be changed at the user level.  */
-  if (perfmonctl(sampling_fd, PFM_START, 0, 0) == -1)
-    panic ("perfmonctl(PFM_START): %s\n", strerror (errno));
+  if (pfm_start(sampling_fd, NULL) == -1)
+    panic ("pfm_start(): %s\n", strerror (errno));
 
   my_cpu = cpu;	/* sampling is active now */
 
@@ -1078,19 +1064,19 @@ profile_cpu (int cpu)
       if (buffer_full)
 	/* read & ignore the overflow  notification message */
 	read (sampling_fd, &msg, sizeof (msg));
-      else if (perfmonctl (sampling_fd, PFM_STOP, 0, 0) == -1)
-	panic ("perfmonctl(PFM_STOP): %s\n", strerror (errno));
+      else if (pfm_stop(sampling_fd) == -1)
+	panic ("pfm_stop(): %s\n", strerror (errno));
 
       process_pmu_samples ();
 
       if (done)
 	break;
 
-      if (perfmonctl (sampling_fd, PFM_RESTART, NULL, 0) < 0)
-	panic ("perfmonctl(PFM_RESTART)\n");
+      if (pfm_restart(sampling_fd) < 0)
+	panic ("pfm_restart()\n");
 
-      if (!buffer_full && perfmonctl (sampling_fd, PFM_START, 0, 0) == -1)
-	panic ("perfmonctl(PFM_START): %s\n", strerror (errno));
+      if (!buffer_full && pfm_start(sampling_fd, NULL) == -1)
+	panic ("pfm_start(): %s\n", strerror (errno));
     }
 
   kallsyms = NULL;
